"""
Allows the user to select an area of the screen that the autofisher
will watch for subtitles. When the `grab()` method is called, it will
open a Tkinter window where the user can click and drag a rectangular
area whose coordinates will then be returned
"""
import sys
import json
import tkinter as tk


def _mouseMove(c, rect, x, y):
    """Moves the lower right corner of a rectangle on
    a Tkinter Canvas to the given (x, y)-coordinates,
    effectively stretching the rectangle

    Arguments:
        c (Tkinter.Canvas)
            Any Tkinter Canvas object
        rect (int)
            The identifier of a rectangle on the canvas,
            returned by canvas.create_rectangle
        x (int)
            Destination x-coordinate
        y (int)
            Destination y-coordinate
    """
    x1, y1, _, _ = c.coords(rect)
    c.coords(rect, x1, y1, x, y)


def grab():
    """Opens a Tkinter window and allows the user to
    click and drag a rectangular selection, then returns
    the coordinates of the selection
    
    Returns:
        list:
            A list of 4 numbers (x1, y1, x2, y2), where
            (x1, y1) is the top left corner and (x2, y2)
            is the lower right corner
    """
    bbox = [0, 0, 0, 0]

    # Set up Tkinter window
    root = tk.Tk()
    root.wait_visibility(root)
    root.attributes('-alpha', 0.1)
    root.attributes('-fullscreen', True)

    # Create Tkinter Canvas and Rect object
    c = tk.Canvas(root)
    c.pack(fill='both', expand=True)
    rect = c.create_rectangle(0, 0, 0, 0, fill="blue")

    def end():
        nonlocal bbox
        bbox = c.coords(rect)
        root.destroy()

    # Set event bindings and start Tkinter main loop
    root.bind("<ButtonPress-1>", lambda e: c.coords(rect, e.x_root, e.y_root, e.x_root, e.y_root))
    root.bind("<B1-Motion>", lambda e: _mouseMove(c, rect, e.x_root, e.y_root))
    root.bind("<ButtonRelease-1>", lambda e: end())
    root.bind("<Escape>", lambda e: end())
    root.mainloop()

    return bbox



"""
Automated Fishing Program for Minecraft

Command Line Usage
^^^^^^^^^^^^^^^^^^
    .. code:: bash

        python -m mc-autofisher [--levdist x] [--tesspath x] [-s|--new-bbox] [...]
    
    **Arguments:**
        ``--levdist x``: Integer (default 5)
            The maximum Levenshtein distance allowed for text matching, accounts for slight
            errors in Tesseract's text recognition, i.e. 'Fishing Gobber' instead of 'Fishing Bobber'
        ``--tesspath x``: String
            The path to your Tesseract installation, usually unnecessary unless you have some special setup
        ``--new-bbox`` or ``-s``
            Tells the program you want to run the screengrabber and record the new bounding box in the config
            file. The screengrabber will run by default if there is not a bounding box already recorded in the
            config file and you have not manually provided coordinates on the command line

    **Other Parameters:**
        If four flagless arguments are present they will be treated as manually-entered
        coordinates of the top left and bottom right corners of the bounding box and will take precedence
        over ``--new-bbox`` and the coordinates in the config file
    
    **Examples:**
        .. code:: bash

            $ python -m mc-autofisher                               # Runs with parameters from config file

            $ python -m mc-autofisher 50 50 100 100                 # Runs with specified bounding box, all other parameters from config file

            $ python -m mc-autofisher --levdist 1                   # Runs with a maximum Levenshtein distance of 1

            $ python -m mc-autofisher --tesspath /home/Desktop      # Why would you install Tesseract on your Desktop?

            $ python -m mc-autofisher --new-bbox                    # Runs the screengrabber first to allow selection of a new bounding box
"""
from .fisher import start
from .screengrabber import grab

__author__ = "Robert Best"
__email__  = "bobbyisbest3@gmail.com"
__version__= "1.0.2"



"""
Automates Minecraft fishing by watching a portion of the screen for the subtitle
text "Fishing Bobber splashes", and double clicking the right mouse button when it
is found, allowing it to reel in and recast the line every time a fish appears.
"""

import sys
import time
import json
import pyscreenshot as ig
import pytesseract as pt
from pynput.keyboard import Key, Listener
from pynput.mouse import Button, Controller
from Levenshtein._levenshtein import distance as levenshtein


def _keyDown(e):
    """Callback for keyboard listener, closes
    the listener if the Enter key is pressed
    """
    if e == Key.enter:
        print('Exiting')
        return False
    return True


def match(screenText, targetText="Fishing Bobber", threshold=5):
    """Takes in some text and iterates through it line by line,
    checking to see if any of the lines contain something similar
    to the target text using Levenshtein distance

    Arguments:
        screenText (str)
            The text to search for a match in
        targetText (`string`, optional)
            The text to search for, defaults to "Fishing Bobber"
        threshold: (`int`, optional)
            The maximum acceptable Levenshtein distance to
            be considered a match, defaults to 5
    
    Returns:
        True if a match was found, False otherwise
    
    Examples:
        >>> match("Fishing Bobber")
        True
        >>> match("Fishing Gobber")
        True
        >>> match("Fishing")
        False 
        >>> match("Some text Fishing Bobber some more text")
        True
    """
    ngram_size = len(targetText.split())
    for line in screenText.splitlines():
        line = line.split()
        for i in range(len(line)-ngram_size+1):
            ngram = " ".join(line[i:i+ngram_size])
            if levenshtein(ngram, targetText) <= threshold:
                return True
    return False

def start(bbox, allowed_error=5, tesspath=""):
    """Takes in bounding box coordinates and begins watching
    that section of the screen for the text "Fishing Bobber splashes",
    double clicking the right mouse button when it sees it to reel in
    the fish and recast the line. Press enter to quit.

    Arguments:
        bbox (list)
            A list of 4 numbers (x1, y1, x2, y2) obtained from the screengrabber
            program, where (x1, y1) is the top left corner of the bounding box
            and (x2, y2) is the lower right corner
        allowed_error (`int`, optional)
            The margin of error (measured in Levenshtein distance) allowed for
            text matching, i.e. 'Fishing Gobber splashes' has an error of 1
        tesspath (`str`, optional)
            The path to your Tesseract installation, leave blank to use default
    """
    if tesspath != "":
        pt.pytesseract.tesseract_path = tesspath

    mouse = Controller()
    listener = Listener(on_press=_keyDown)
    listener.start()

    timeout = 45 # Fish always appears within 45 seconds, if we wait longer than this then recast the line
    cast_time = time.monotonic()
    while listener.is_alive():
        # Grab portion of screen defined in config file & send it to tesseract
        screen = ig.grab(bbox=bbox)
        screen_text = pt.image_to_string(screen)
        if match(screen_text, threshold=allowed_error) or time.monotonic() - cast_time > timeout:
            # Either fish was found or timeout was exceeded, reel in the line and cast it again
            mouse.click(Button.right, 2)
            cast_time = time.monotonic()
        time.sleep(0.25)

    listener.join()
